package kosoft.utils.redirectstdio;

// redirectstdio library to redirect java stdout/stderr to timestamped & size rolled file. 
// Copyright (C) 2000 Mark O'Donohue [ mark.odonohue@gmail.com] 

// This library is free software; you can redistribute it and/or modify it under the terms of the GNU 
// Lesser General Public License as published by the Free Software Foundation; either version 2.1 of 
// the License, or (at your option) any later version.

// This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
// even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
// Lesser General Public License for more details.


import java.io.*;
import java.text.SimpleDateFormat;
import java.util.*;
import java.net.*;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

public class RedirectStdio {

    static {
        //System.out.println("Attempting to force UTC time...");
        //System.setProperty("user.timezone","UTC");
        //TimeZone.setDefault(new SimpleTimeZone(0, "UTC"));
        //System.out.println("Timezone: " + TimeZone.getDefault().getID()); 

    }
    
    static RedirectStdio instance = new RedirectStdio();
    
    static boolean redirectedStdio = false;
    public static boolean getIsRedirectedStdio() { return redirectedStdio; } 

    static RuntimeProperties properties = new RuntimeProperties();

    public void loadProperties(String filename) {
        
        File file = new File(filename);
        
        if (!file.exists()) {
            System.err.println("Properties file does not exist" + file.getName());
            return; 
        }

        // So load it.
        properties.load(filename);
        if (isBasicDebug()) {
            System.out.println("Load from properties file : " + file.getName());
        }
        
        // If we have a properties file, then startup bg thread program to check if it changes.
        loadPropFile = file;
        startBgCheckThread();
    }

    public void loadChangedProperties(File file ) {
        
        // Read new properties
        Properties newProperties = new Properties ();
        
        if (!file.exists()) {
            //throw new RuntimeException("Missing required property file : " + file.getName());
            System.out.println("Missing property file : " + file.getName());
            return;
        }
        
        FileInputStream input = null;
        try { 
            input = new FileInputStream(file);
            newProperties.load(input);
        }
        catch (FileNotFoundException e) {
            if (isReportCheckError()) {
                //throw new RuntimeException("FileNotFoundException: " + file.getName());
                System.out.println("Missing property file : " + file.getName());
            }
            return;
            
        }
        catch (IOException e) {
            if (isReportCheckError()) {
                //throw new RuntimeException("IOException reading:" + propertyFileName);            
                System.out.println("IOException reading:" + file.getName());
            }
            return;
        }
        finally {
            try { 
                if (input != null) input.close();
            }
            catch (IOException e) {}
        }

        
        // Transfer new properties to the existing set. 
        // retains old values.
        for(String key: newProperties.stringPropertyNames()) {
            String value = newProperties.getProperty(key, null);
            
            properties.getProperties().setProperty(key, value);
        }
        
        if (isBasicDebug()) {
            System.out.println("Load from modified properties file : " + file.getName());
        }
        
        runCheckRedirect();
    }
    

    long delayTime = -1;
    File loadPropFile = null;

    long lastModifiedDelayTimeFile = -1;
    long nextCheckTime = -1;
    long nextCheckDelayTime = 30000;

    long lastReportCheckError = 0;  
    
    boolean isReportCheckError() { 
        if (lastReportCheckError++ == 0) {
            return true; 
        }
        
        if (lastReportCheckError > 3 * 60) {
            lastReportCheckError = -1;
        }
        return false;
    }

    void checkForFileUpdate() { 

        if (isBgCheckDebug()) {
            if (loadPropFile == null) {
                System.out.println("Check for change in file : " +  "null");
            }
            else {
                try {
                    System.out.println("Check for change in file : " + loadPropFile.getName());
                    //System.out.println("Check for change in file : " + loadPropFile.getAbsolutePath());
                    System.out.println("Path to file is  : " + loadPropFile.getCanonicalPath());
                }
                catch (IOException e) {}
            }
        }

        // If file does not exist. 
        if (loadPropFile == null) return;

        long currentTime = System.currentTimeMillis();

        if (nextCheckTime > currentTime) {
            return;
        }

        // Even if fail below, still only check every X time - otherwise would add CPU churn. 
        nextCheckTime = System.currentTimeMillis() + nextCheckDelayTime;

        if  (!loadPropFile.exists()) { // Does not exist. 
            System.out.println("Unable to find file : " + loadPropFile.getName());
            return;
        }

        long newModifyTime = loadPropFile.lastModified();

        if (lastModifiedDelayTimeFile == newModifyTime ) { // Unchanged since last read.
            return;
        }

        lastModifiedDelayTimeFile = newModifyTime;  // We dont care if fails after this we still only what to check again if changed. 

        loadChangedProperties(loadPropFile);


    }

    boolean isBasicDebug() { return true; } 

    boolean isBgCheckDebug() { return true; } 
    
    void runDelayCheck() {
        
        if (isBasicDebug()) {
            System.out.println("Start Bg Delay Change check : ");
        }
        
        while (true) {
            try {
                if (isBgCheckDebug()) {
                    System.out.println("Start bg checkForFileUpdate");
                }
                
                checkForFileUpdate();
                if (isBgCheckDebug()) {
                    System.out.println("Finish bg checkForFileUpdate : (sleep for : " + nextCheckDelayTime + " ms)");
                }
                
                Thread.sleep(nextCheckDelayTime);
            }
            catch (InterruptedException e) {}
            catch (Throwable e) {
                System.out.println(e.getMessage());
                e.printStackTrace();
            }
        }
        //if (isBasicDebug()) {
        //  System.out.println("Exiting Bg Delay Change check : ");
        //}
        
        
    }
    
    Thread bgThread = null; 
    void startBgCheckThread() { 
        bgThread = new Thread() {
            public void run() {
                try { 
                    runDelayCheck();
                }
                finally {
                    bgThread = null;
                }
            }
        };

        bgThread.setDaemon(true);       
        bgThread.start();
    }
    
    // <Cmd> prefix=C:\Logging_ rolloverSize=5000

    // Some static methods that redirect to instance methods, the internal ones are synchronized, just in case
    // two people request redirect at the same time.    
    
    public static void doRedirectCmd(String prefix) {
        instance.doRedirectCmdImpl(prefix);
    }

    public static void checkRedirect() {
        instance.checkRedirectImpl();
    }
    
    public static void clearRedirect() {
        instance.clearRedirectImpl();
    }
    
    
    private synchronized void doRedirectCmdImpl(String prefix) {            
        
        if (prefix != null) {
            
            StringTokenizer tk = new StringTokenizer(prefix, ",");
            while (tk.hasMoreElements()) {
                String pair = tk.nextToken();
                if (pair.indexOf("=") != -1) {
                    StringTokenizer tk2 = new StringTokenizer(pair,"=");
                    if (tk2.countTokens() == 2) {
                        String key = tk2.nextToken();
                        String value = tk2.nextToken();
                        value = URLDecoder.decode(value);
                        properties.getProperties().setProperty(key, value);
                        //System.out.println("Setting " + key + "=" + value);
                    }
                }
                else if (prefix.endsWith(".properties")) {
                    System.out.println("loading properties file " + pair);
                    loadProperties(pair);
                }
                else {
                    System.out.println("Uknown option : " + pair);
                }
            }
        }   

        runCheckRedirect();
    }

    void runCheckRedirect() { 
        
        boolean enable = properties.getPropertyAsBoolean("enable", true);       
        nextCheckDelayTime = properties.getPropertyAsInt("nextCheckDelayTime", 30000);
        
        if (isBasicDebug()) {
            System.out.println("Redirect Enable status is : " + enable);
        }
        
        if (enable) {          
            checkRedirectImpl(); 
        }
        else  {
            clearRedirectImpl();
        }

    }

    static PrintStream oldErr = null;
    static PrintStream oldOut = null;

    private synchronized void clearRedirectImpl() {
        if (redirectedStdio) {
            Date timestamp  = new Date();

            System.out.println("stopping Redirect stdout: " + timestamp.toString());
            System.err.println("stopping Redirect stderr: " + timestamp.toString());            

            PrintStream redirectOut = System.out;
            PrintStream redirectErr = System.err;

            System.setErr(oldErr);
            System.setOut(oldOut);

            // Close the redirect stream, with test to make sure its not real stdout.
            if (redirectOut != oldOut) { 
                redirectOut.flush();
                redirectOut.close();
            } 

            // Close rediret stdErr, with test that it is not the real stderr, or if redirect err file was 
            // to same file as stdout redirect (which is usually the case).
            if (redirectErr != oldErr  && redirectErr != redirectOut) { 
                redirectErr.flush();
                redirectErr.close();
            }


            oldErr = null;
            oldOut = null;

            redirectedStdio = false;
        }
    }


    
    String convToHexString(byte[] bytes) {
        
        StringBuilder sb = new StringBuilder(2 * bytes.length);
        for (byte b : bytes) {
            sb.append("0123456789ABCDEF".charAt((b & 0xF0) >> 4));
            sb.append("0123456789ABCDEF".charAt((b & 0x0F)));
        }
        String hex = sb.toString();
        
        return hex;
    }
    
    String currentSettingsHash = null;
    
    private synchronized void checkRedirectImpl() {

        String prefix = properties.getProperty("prefix", "");
        long rolloverSize = properties.getPropertyAsLong("rolloverSize", 4000);
        rolloverSize *= 1000; // make it in K.

        boolean isRedirectStdoutStderrToSameFile = properties.getPropertyAsBoolean("combineStdErr", true);
        int numLogsToRetain = properties.getPropertyAsInt("numLogsToRetain", 10);

        
        boolean changedSettings = false; 
        
        try {
            MessageDigest  newHash = MessageDigest.getInstance("SHA-256");
            
            newHash.update(prefix.getBytes());
            newHash.update(Long.toString(rolloverSize).getBytes());
            newHash.update(Boolean.toString(isRedirectStdoutStderrToSameFile).getBytes());          
            newHash.update(Long.toString(numLogsToRetain).getBytes());
            byte[] digest = newHash.digest();
            
            String newSettingsHash = convToHexString(digest);
            
            if (currentSettingsHash == null || !currentSettingsHash.equals(newSettingsHash)) {
                changedSettings = true;
                currentSettingsHash = newSettingsHash;
            }

            
        } catch (NoSuchAlgorithmException e) {
        }
        
        boolean debug = properties.getPropertyAsBoolean("debug", false);
        
        if (debug) {
            System.out.println("Setting RedirectStdio : ");
            System.out.println("prefix          =" + prefix);
            System.out.println("rolloverSize    =" + rolloverSize);
            System.out.println("combineStderr   =" + isRedirectStdoutStderrToSameFile);
            System.out.println("numLogsToRetain =" + numLogsToRetain);
            System.out.println("debug           =" + debug);
        }

        // If we have changed the settings and are already redirecting then
        // we could be smarter, so we dont loose any redirect data should 
        if (redirectedStdio && changedSettings) {
            clearRedirectImpl();
        }
        
        if (!redirectedStdio) {

            oldErr = System.err;
            oldOut = System.out;

            RollingOutputStream rosStdout = new RollingOutputStream(prefix + "stdout", rolloverSize);
            rosStdout.setNumLogsToRetain(numLogsToRetain);
            rosStdout.setDebug(debug);
            
            PrintStream newOut = new PrintStream(rosStdout, true);
            PrintStream newErr = null;
            if (isRedirectStdoutStderrToSameFile) {
                newErr = newOut;
            }
            else {
                RollingOutputStream rosStderr = new RollingOutputStream(prefix + "stdout", rolloverSize);
                rosStderr.setNumLogsToRetain(numLogsToRetain);
                newErr = new PrintStream(rosStderr, true);
            }
            System.setErr(newErr);
            System.setOut(newOut);

            Date timestamp  = new Date();
            
            if (debug) {
                System.out.println("Setting RedirectStdio : ");
                System.out.println("prefix          =" + prefix);
                System.out.println("rolloverSize    =" + rolloverSize);
                System.out.println("combineStderr   =" + isRedirectStdoutStderrToSameFile);
                System.out.println("numLogsToRetain =" + numLogsToRetain);
                System.out.println("debug           =" + debug);
            }
            
            System.out.println("Started Redirect stdout: " + timestamp.toString());
            if (!isRedirectStdoutStderrToSameFile) {
                System.err.println("started Redirect stderr: " + timestamp.toString());
            }

            redirectedStdio = true;
        }
    }


    public static void main(String argv[]) {
        runTest();
    }

    public static void runTest() {

        //properties.getProperties().setProperty("rolloverSize", "5000");
        properties.getProperties().setProperty("rolloverSize", "2");
        properties.getProperties().setProperty("numLogsToRetain", "5");
        //checkRedirect();

        doRedirectCmd("dist/redirectstdio.properties");
        
        for (int i=0; i< 100000; i++) {
            System.out.println("test1 " + i);
            System.out.flush();
            System.err.println("test2 " + i);
            System.err.flush();

            try { 
                Thread.sleep(100);
            }
            catch (InterruptedException e) {}
            
        }
        
        try { 
            Thread.sleep(300);
        }
        catch (InterruptedException e) {}
    
    }

}
